use std::fs;
use std::io::Read;
use std::path::Path;
use regex::Regex;
use clap::Parser;

#[derive(Debug)]
struct CFile {
    path: String,
    filename: String,
    headers: Vec<String>,
}

#[derive(Parser, Debug)]
struct Args {
    /// Replace default make output with nice build messages
    #[arg(short, long)]
    pretty: bool,
    
    #[arg(default_value = "./build")]
    build_dir: String,
}

fn parse_c_file(path: &Path, re: &Regex) -> CFile {
    let strpath = path.to_str().unwrap();
    let filename = path.file_stem().unwrap().to_str().unwrap();

    let mut file_handle = fs::File::open(path).unwrap();
    let mut contents = String::new();
    file_handle.read_to_string(&mut contents).unwrap();

    let mut headers: Vec<String> = Vec::new();
    let caps = re.captures_iter(&contents);
    for (_, [include_file]) in caps.map(|c| c.extract()) {
        headers.push(path.parent().unwrap()
                         .join(Path::new(include_file))
                         .to_str().unwrap()
                         .to_string());
    }

    return CFile { path: strpath.to_string(), filename: filename.to_string(), headers }

}

fn recurse_into_dir<F: FnMut(fs::DirEntry)>(path: &Path, file_callback: &mut F) {
    let files = fs::read_dir(path).unwrap();
    for file in files {
        let file = file.unwrap();
        let meta = file.metadata().unwrap();
        let is_dir = meta.is_dir();
        let is_file = meta.is_file();
        let path = file.path();
        let name = file.file_name();

        // ignore hidden files
        if name.to_str().unwrap().chars().nth(0).unwrap() == '.' {
            continue;
        }

        if is_dir {
            recurse_into_dir(&path, file_callback);
        } else if is_file {
            file_callback(file);
        }
    }
}

fn main() {
    let mut project_files: Vec<CFile> = Vec::new();
    let includes_regex = Regex::new(r#"#include\s+"(.*?)""#).unwrap();

    let args = Args::parse();
    let build_dir = Path::new(&args.build_dir);
    let pretty = args.pretty;

    let obj_dir = build_dir.join("obj");

    recurse_into_dir(Path::new("./"), &mut |file| {
        let path = file.path();
        match path.extension() {
            Some(v) => {
                let v = v.to_str().unwrap();
                if v != "c" {
                    return;
                }
                project_files.push(parse_c_file(&path, &includes_regex));
            },
            None => {
                return;
            },
        }
    });

    for file in project_files.iter() {
        let mut build_rule: String = "#=mgen_rule=#\n".to_owned();
        
        let objname = file.filename.to_owned() + ".o";
        let objpath = obj_dir.join(&objname);
        let objpath = objpath.to_str().unwrap();

        build_rule.push_str(&format!("{}: {}", objpath, file.path));
        for header in file.headers.iter() {
            build_rule.push_str(&format!(" {}", header));
        }

        build_rule.push_str(&format!("\n\t$(CC) $(CFLAGS) -o {} -c {}\n",  objpath, file.path ));

        println!("'{}'", build_rule);
    }

}
